from unittest import TestCase

from sekg.text.util import CodeTextPreprocessor


class TestCodeTextPreprocessor(TestCase):
    def test_clean_html_text(self):
        html_text = """
<!doctype html>
<html>
  <head>
    <!-- MathJax -->
    <script type="text/javascript"
      src="http://cdn.mathjax.org/mathjax/latest/MathJax.js?config=TeX-AMS-MML_HTMLorMML">
    </script>
    <meta charset="utf-8">
    <meta http-equiv="X-UA-Compatible" content="chrome=1">
    <title>
      Caffe | Bias Layer
    </title>

    <link rel="icon" type="image/png" href="/images/caffeine-icon.png">

    <link rel="stylesheet" href="/stylesheets/reset.css">
    <link rel="stylesheet" href="/stylesheets/styles.css">
    <link rel="stylesheet" href="/stylesheets/pygment_trac.css">

    <meta name="viewport" content="width=device-width, initial-scale=1, user-scalable=no">
    <!--[if lt IE 9]>
    <script src="//html5shiv.googlecode.com/svn/trunk/html5.js"></script>
    <![endif]-->
  </head>
  <body>
  <script>
    (function(i,s,o,g,r,a,m){i['GoogleAnalyticsObject']=r;i[r]=i[r]||function(){
    (i[r].q=i[r].q||[]).push(arguments)},i[r].l=1*new Date();a=s.createElement(o),
    m=s.getElementsByTagName(o)[0];a.async=1;a.src=g;m.parentNode.insertBefore(a,m)
    })(window,document,'script','//www.google-analytics.com/analytics.js','ga');

    ga('create', 'UA-46255508-1', 'daggerfs.com');
    ga('send', 'pageview');
  </script>
    <div class="wrapper">
      <header>
        <h1 class="header"><a href="/">Caffe</a></h1>
        <p class="header">
          Deep learning framework by <a class="header name" href="http://bair.berkeley.edu/">BAIR</a>
        </p>
        <p class="header">
          Created by
          <br>
          <a class="header name" href="http://daggerfs.com/">Yangqing Jia</a>
          <br>
          Lead Developer
          <br>
          <a class="header name" href="http://imaginarynumber.net/">Evan Shelhamer</a>
        <ul>
          <li>
            <a class="buttons github" href="https://github.com/BVLC/caffe">View On GitHub</a>
          </li>
        </ul>
      </header>
      <section>

      <h1 id="bias-layer">Bias Layer</h1>

<ul>
  <li>Layer type: <code class="highlighter-rouge">Bias</code></li>
  <li><a href="http://caffe.berkeleyvision.org/doxygen/classcaffe_1_1BiasLayer.html">Doxygen Documentation</a></li>
  <li>Header: <a href="https://github.com/BVLC/caffe/blob/master/include/caffe/layers/bias_layer.hpp"><code class="highlighter-rouge">./include/caffe/layers/bias_layer.hpp</code></a></li>
  <li>CPU implementation: <a href="https://github.com/BVLC/caffe/blob/master/src/caffe/layers/bias_layer.cpp"><code class="highlighter-rouge">./src/caffe/layers/bias_layer.cpp</code></a></li>
  <li>CUDA GPU implementation: <a href="https://github.com/BVLC/caffe/blob/master/src/caffe/layers/bias_layer.cu"><code class="highlighter-rouge">./src/caffe/layers/bias_layer.cu</code></a></li>
</ul>

<h2 id="parameters">Parameters</h2>
<ul>
  <li>Parameters (<code class="highlighter-rouge">BiasParameter bias_param</code>)</li>
  <li>From <a href="https://github.com/BVLC/caffe/blob/master/src/caffe/proto/caffe.proto"><code class="highlighter-rouge">./src/caffe/proto/caffe.proto</code></a>):</li>
</ul>

<figure class="highlight"><pre><code class="language-protobuf" data-lang="protobuf"><span class="kd">message</span> <span class="nc">BiasParameter</span> <span class="p">{</span>
  <span class="c1">// The first axis of bottom[0] (the first input Blob) along which to apply
</span>  <span class="c1">// bottom[1] (the second input Blob).  May be negative to index from the end
</span>  <span class="c1">// (e.g., -1 for the last axis).
</span>  <span class="c1">//
</span>  <span class="c1">// For example, if bottom[0] is 4D with shape 100x3x40x60, the output
</span>  <span class="c1">// top[0] will have the same shape, and bottom[1] may have any of the
</span>  <span class="c1">// following shapes (for the given value of axis):
</span>  <span class="c1">//    (axis == 0 == -4) 100; 100x3; 100x3x40; 100x3x40x60
</span>  <span class="c1">//    (axis == 1 == -3)          3;     3x40;     3x40x60
</span>  <span class="c1">//    (axis == 2 == -2)                   40;       40x60
</span>  <span class="c1">//    (axis == 3 == -1)                                60
</span>  <span class="c1">// Furthermore, bottom[1] may have the empty shape (regardless of the value of
</span>  <span class="c1">// "axis") -- a scalar bias.
</span>  <span class="k">optional</span> <span class="kt">int32</span> <span class="na">axis</span> <span class="o">=</span> <span class="mi">1</span> <span class="p">[</span><span class="k">default</span> <span class="o">=</span> <span class="mi">1</span><span class="p">];</span>

  <span class="c1">// (num_axes is ignored unless just one bottom is given and the bias is
</span>  <span class="c1">// a learned parameter of the layer.  Otherwise, num_axes is determined by the
</span>  <span class="c1">// number of axes by the second bottom.)
</span>  <span class="c1">// The number of axes of the input (bottom[0]) covered by the bias
</span>  <span class="c1">// parameter, or -1 to cover all axes of bottom[0] starting from `axis`.
</span>  <span class="c1">// Set num_axes := 0, to add a zero-axis Blob: a scalar.
</span>  <span class="k">optional</span> <span class="kt">int32</span> <span class="na">num_axes</span> <span class="o">=</span> <span class="mi">2</span> <span class="p">[</span><span class="k">default</span> <span class="o">=</span> <span class="mi">1</span><span class="p">];</span>

  <span class="c1">// (filler is ignored unless just one bottom is given and the bias is
</span>  <span class="c1">// a learned parameter of the layer.)
</span>  <span class="c1">// The initialization for the learned bias parameter.
</span>  <span class="c1">// Default is the zero (0) initialization, resulting in the BiasLayer
</span>  <span class="c1">// initially performing the identity operation.
</span>  <span class="k">optional</span> <span class="n">FillerParameter</span> <span class="na">filler</span> <span class="o">=</span> <span class="mi">3</span><span class="p">;</span>
<span class="p">}</span></code></pre></figure>



      </section>
  </div>
  </body>
</html>
"""
        preprocessor = CodeTextPreprocessor()
        clean = preprocessor.clean_html_text(html_text=html_text)
        print(clean)



from gensim.models import Word2Vec
from gensim.models.keyedvectors import Word2VecKeyedVectors

"""
    Here is some pretrain word2vec:
    https://github.com/3Top/word2vec-api

"""


class TunedWord2VecTrainer:
    @staticmethod
    def tune(corpus, pretrain_w2v_path, pretrain_binary=True, window=5, min_count=1, size=100, iter=1, alpha=0.0001):
        """
        tune the pretrained word2vec
        :param corpus:
        :param pretrain_w2v_path:
        :param pretrain_binary:
        :param window:
        :param min_count:
        :param size:
        :param iter: the iter can't be too much, otherwise the pretrained Word2Vec won't be helpful.
        :param alpha: the learning rate, the alpha. alpha the can't be too high.
        :return:
        """
        pretrained_word2vec_model = Word2VecKeyedVectors.load_word2vec_format(
            pretrain_w2v_path,
            binary=pretrain_binary)

        w2v = Word2Vec(window=window, min_count=min_count, size=size, iter=iter, alpha=alpha)
        w2v.build_vocab(corpus)
        training_examples_count = w2v.corpus_count
        w2v.build_vocab([list(pretrained_word2vec_model.vocab.keys())], update=True)

        w2v.intersect_word2vec_format(pretrain_w2v_path, binary=pretrain_binary, lockf=1.0)
        w2v.train(corpus, total_examples=training_examples_count, epochs=w2v.epochs)
        return w2v



import numpy as np
from gensim.corpora import Dictionary
from gensim.models import TfidfModel
from gensim.models.keyedvectors import Word2VecKeyedVectors

from sekg.graph.exporter.graph_data import GraphData
from sekg.graph.util.name_searcher import KGNameSearcher
from sekg.ir.doc.wrapper import PreprocessMultiFieldDocumentCollection


class LoadUtil:
    DEFAULT_EMBEDDING_SIZE = 100

    @staticmethod
    def load_graph_data(graph_data_path, graph_data: GraphData):
        if graph_data is None:
            if graph_data_path is not None:
                print("graph_data is given, loading")

                return GraphData.load(graph_data_path)
            else:
                print("graph_data is not given, training")
                # todo: implement training the GraphData here, currently is only load a pretrain KGNameSearcher  model
        else:
            return graph_data

    @staticmethod
    def load_kg_name_searcher(kg_name_searcher: KGNameSearcher, kg_name_searcher_path):
        if kg_name_searcher:
            return kg_name_searcher

        if kg_name_searcher is None:

            if kg_name_searcher_path is not None:
                print("pretrain_search_model_path is given, loading")

                return KGNameSearcher.load(kg_name_searcher_path)
            else:
                print("pretrain_search_model_path is not given, training, but the training is not implemented yet")
                # todo: implement training the KGNameSearcher here, currently is only load a pretrain KGNameSearcher  model

    @staticmethod
    def load_tfidf_model(tfidf_model_path, tfidf_model: TfidfModel, dict: Dictionary, corpus):
        if tfidf_model:
            return tfidf_model

        if tfidf_model_path is not None:
            print("pretrain tfidf path is given, loading")
            return TfidfModel.load(tfidf_model_path)

        else:
            print("tfidf Training...")
            tfidf_model = TfidfModel(corpus=corpus, id2word=dict)
            print("tfidf Train complete")
            return tfidf_model

    @staticmethod
    def get_unknown_vector(embedding_size):
        unknwon_vector = np.zeros(embedding_size)
        unknwon_vector[0] = 1e-07
        return unknwon_vector

    @classmethod
    def load_node2vec_for_doc_collection(cls, preprocess_document_collection: PreprocessMultiFieldDocumentCollection,
                                         pretrain_node2vec_path=None,
                                         embedding_size=DEFAULT_EMBEDDING_SIZE,

                                         ):
        """
        train the node2vec model
        :param preprocess_document_collection:
        :param pretrain_node2vec_path:
        :param embedding_size:
        :return:
        """

        not_exist_vector = LoadUtil.get_unknown_vector(embedding_size)

        if pretrain_node2vec_path is not None:
            print("pretrained node2vec path path is given, loading")
            full_node2vec_model = Word2VecKeyedVectors.load(pretrain_node2vec_path)
            new_node2vec_model = Word2VecKeyedVectors(vector_size=embedding_size)

            doc_id_str_list = []
            vector_list = []

            invalid_doc_id_count = 0
            doc_list = preprocess_document_collection.get_all_preprocess_document_list()
            for doc in doc_list:
                doc_id_str = str(doc.get_document_id())
                if doc_id_str not in full_node2vec_model.vocab:
                    vector = not_exist_vector
                    invalid_doc_id_count = invalid_doc_id_count + 1
                else:
                    vector = full_node2vec_model[doc_id_str]

                doc_id_str_list.append(doc_id_str)
                vector_list.append(vector)

            if len(doc_id_str_list) != len(set(doc_id_str_list)):
                raise Exception("error when filter the node2vec, duplicate doc id in input doc collection!")

            new_node2vec_model.add(entities=doc_id_str_list, weights=vector_list, replace=True)
            print("full node2vec size=%d, invald id=%d new_size=%d" % (
                len(full_node2vec_model.vocab), invalid_doc_id_count, len(new_node2vec_model.vocab)))

            return new_node2vec_model

        else:
            print("pretrained node2vec path is not given, training")

            # print("Word2Vec Training...")
            # # using the CBOW model of word2vec, because we don't use to predict
            # w2v_model = gensim.models.Word2Vec(sentences=corpus_clean_text, size=embedding_size, min_count=1)
            # print("Word2Vec Train complete")
            # self.node2vec_model = w2v_model.wv
            # todo: implement training the node2vec here, currently is only load a pretrain node2ve  model



from unittest import TestCase

from sekg.util.code import CodeElementNameUtil


class TestCodeElementNameUtil(TestCase):
    def test_uncamelize(self):
        util = CodeElementNameUtil()
        name_list = [
            ("java.util.List", "java.util. List"),
            ("java.util.ArrayList", "java.util. Array List"),
            ("ArrayList", "Array List"),
            ("FPath1", "F Path1"),
            ("FilePath1", "File Path1"),

        ]
        for name, right_name in name_list:
            self.assertEqual(right_name, util.uncamelize(name))

    def test_uncamelize_by_stemming(self):
        util = CodeElementNameUtil()
        name_list = [
            ("java.util.List", "java.util. List"),
            ("java.util.ArrayList", "java.util. Array List"),
            ("ArrayList", "Array List"),
            ("FPath1", "F Path1"),
            ("FilePath1", "File Path1"),

        ]
        for name, right_name in name_list:
            self.assertEqual(right_name, util.uncamelize_by_stemming(name))
    def test_uncamelize_from_simple_name(self):
        util = CodeElementNameUtil()
        name_list = [
            ("java.util.List", "List"),
            ("java.util.ArrayList", "Array List"),
            ("ArrayList", "Array List"),
            ("FPath1", "F Path1"),
            ("FilePath1", "File Path1"),

        ]
        for name, right_name in name_list:
            self.assertEqual(right_name, util.uncamelize_from_simple_name(name))


import re

import spacy


class SpacyNLPFactory:
    """

    """
    __domain_extractor_nlp = None
    __identifier_extractor_nlp = None
    __simple_nlp = None

    @classmethod
    def create_spacy_nlp_for_domain_extractor(clss):
        """
        load a spacy nlp pipeline for extract domain entity and relations
        :return:
        """
        if clss.__domain_extractor_nlp is not None:
            return clss.__domain_extractor_nlp

        # todo: fix this, write a class as Spacy Component
        nlp = spacy.load("en")
        id_re = re.compile(r"id|ID|Id")

        prefix_re = spacy.util.compile_prefix_regex(nlp.Defaults.prefixes)
        infix_re = spacy.util.compile_infix_regex(nlp.Defaults.infixes)
        suffix_re = spacy.util.compile_suffix_regex(nlp.Defaults.suffixes)
        nlp.tokenizer = spacy.tokenizer.Tokenizer(nlp.vocab, prefix_search=prefix_re.search,
                                                  infix_finditer=infix_re.finditer,
                                                  suffix_search=suffix_re.search, token_match=id_re.match)

        clss.__domain_extractor_nlp = nlp
        return nlp

    @classmethod
    def create_spacy_nlp_for_identifier_extractor(clss):
        """
        load a spacy nlp pipeline for extract domain entity and relations
        :return:
        """
        if clss.__identifier_extractor_nlp is not None:
            return clss.__identifier_extractor_nlp

        # todo: fix this, write a class as Spacy Component
        nlp = spacy.load("en")
        hyphen_re = re.compile(r"[A-Za-z\d]+-[A-Za-z\d]+|'[a-z]+|''")

        prefix_re = spacy.util.compile_prefix_regex(nlp.Defaults.prefixes)
        infix_re = spacy.util.compile_infix_regex(nlp.Defaults.infixes)
        suffix_re = spacy.util.compile_suffix_regex(nlp.Defaults.suffixes)
        nlp.tokenizer = spacy.tokenizer.Tokenizer(nlp.vocab, prefix_search=prefix_re.search,
                                                  infix_finditer=infix_re.finditer,
                                                  suffix_search=suffix_re.search, token_match=hyphen_re.match)

        clss.__identifier_extractor_nlp = nlp
        return nlp

    @classmethod
    def create_simple_nlp_pipeline(clss):
        """
        create a simple nlp pipeline, without NER and dependency parser, could tokenize and pos,lemma, will be very fast.        :return:
        """
        if clss.__simple_nlp is not None:
            return clss.__simple_nlp

        NLP = spacy.load('en', disable=["ner", "parser"])
        hyphen_re = re.compile(r"[A-Za-z\d]+-[A-Za-z\d]+|'[a-z]+|''")
        prefix_re = spacy.util.compile_prefix_regex(NLP.Defaults.prefixes)
        infix_re = spacy.util.compile_infix_regex(NLP.Defaults.infixes)
        suffix_re = spacy.util.compile_suffix_regex(NLP.Defaults.suffixes)
        NLP.tokenizer = spacy.tokenizer.Tokenizer(NLP.vocab, prefix_search=prefix_re.search,
                                                  infix_finditer=infix_re.finditer,
                                                  suffix_search=suffix_re.search, token_match=hyphen_re.match)

        clss.__simple_nlp = NLP
        return NLP



import os

"""
this file contain some util method for file

"""
# todo: complete this method
class DirUtil:
    def __init__(self):
        pass

    @staticmethod
    def create_file_dir(dir_path):
        is_exits = os.path.exists(dir_path)
        if not is_exits:
            os.makedirs(dir_path)
            print("the output dir %r is not exist, creating" % dir_path)
        else:
            print("the output dir %r is existing, not creating" % dir_path)



import re


class CodeElementNameUtil:
    PATTERN_2_4 = re.compile(r'([A-Za-z])([24])([A-CE-Za-ce-z])')
    PATTERN_split = re.compile(r'([A-Z]+)([A-Z][a-z0-9]+)')
    PATTERN_split_num = re.compile(r'([0-9]?[A-Z]+)')

    def get_simple_name_with_parent(self, name):
        if not name:
            return None
        team_name = name.split("(")[0]
        split_names = team_name.split(".")
        if len(split_names) <= 1:
            return split_names[-1]

        child = split_names[-1].strip()
        parent = split_names[-2].strip()

        return parent + "." + child

    def simplify(self, name):
        """
        get the simple name for class, method, field, eg. java.util.ArrayList->ArrayList
        :param name:
        :return:
        """
        if not name:
            return None
        team_name = name.split("(")[0]
        simple_name = team_name.split(".")[-1].strip()

        return simple_name

    def uncamelize_from_simple_name(self, name):
        """
        uncamel from simple name of one name, rg. java.util.ArrayList->Array List
        :param name:
        :return:
        """
        if not name:
            return None
        simple_name = self.simplify(name)

        return self.uncamelize(simple_name)

    def uncamelize(self, name):
        """
        uncamel one name
        :param name: the camel styple name(include underline)
        :return:
        """
        if not name:
            return None
        # sub = re.sub(r'([A-Za-z])([24])([A-CE-Za-ce-z])', r'\1 \2 \3', name).strip()
        sub = re.sub(self.PATTERN_2_4, r'\1 \2 \3', name).strip()
        sub = re.sub(r'_', " ", sub)
        # sub = re.sub(r'([A-Z]+)([A-Z][a-z0-9]+)', r'\1 \2', sub)
        sub = re.sub(self.PATTERN_split, r'\1 \2', sub)
        # sub = re.sub(r'([0-9]?[A-Z]+)', r' \1', sub)
        sub = re.sub(self.PATTERN_split_num, r' \1', sub)
        sub = re.sub(r'\s+', " ", sub).strip()
        return sub

    def uncamelize_by_stemming(self, name):
        """
        uncamelzie the name and remove last num, eg. Student1->Student, JavaParser3->Java Parser
        :param name:
        :return:
        """
        # todo: improve this method to fix more situation, has some error for Path1->Path
        name = self.uncamelize(name)
        if not name:
            return None
        sub = re.sub(r'([0-9]+)$', '', name)
        return sub

    def generate_aliases(self, qualified_name, include_simple_parent_name=False):
        if not qualified_name:
            return []

        simple_name = self.simplify(qualified_name)
        separate_name = self.uncamelize_from_simple_name(simple_name)

        name_list = [qualified_name, simple_name, separate_name, ]

        if include_simple_parent_name:
            name_list.append(self.get_simple_name_with_parent(qualified_name))

        name_list = [name for name in name_list if name]

        return list(set(name_list))



#!/usr/bin/env python
# -*- encoding: utf-8 -*-

# Copyright 2011-2018, Nigel Small
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

__all__ = ["__author__", "__email__", "__license__", "__package__", "__version__",
           ]

__author__ = "FDSEKG"
__email__ = "lmwtclmwtc@outlook.com"
__license__ = "Apache License, Version 2.0"
__package__ = "sekg"
__version__ = "0.4.31"



import networkx as nx
from gensim.models.keyedvectors import Word2VecKeyedVectors

from sekg.graph.el.base import EntryPointLinker, EntityLinkResult
from sekg.graph.util.name_searcher import KGNameSearcher
from sekg.ir.models.base import DocumentSimModel
from sekg.ir.preprocessor.base import Preprocessor


class BestSeqSelector:

    def __init__(self, ):
        self.keyword2candidate_map = {}
        self.G = nx.Graph()
        self.best_selection = []
        self.best_score = 0.0

    def get_candidate_ids(self):
        result = set([])
        for keyword, ids in self.keyword2candidate_map.items():
            result = result | ids
        return list(result)

    def get_keyword_list(self):
        return list(self.keyword2candidate_map.keys())

    def add_candidate_for_keyword(self, keyword, candidate_node_id, score):
        if keyword not in self.keyword2candidate_map:
            self.keyword2candidate_map[keyword] = set([])

        self.keyword2candidate_map[keyword].add(candidate_node_id)
        self.add_candidate_entity_score(node_id=candidate_node_id, score=score)

    def add_candidate_entity_score(self, node_id, score):
        self.G.add_node(node_id, score=score)

    def add_pair_score(self, start_id, end_id, score):
        self.G.add_edge(start_id, end_id, score=score)

    def search_best_combination(self):
        self.chosen([], score=0.0, current_chosen_keyword_index=0)

        return self.get_best_selection()

    def get_best_score(self):
        return self.best_score

    def get_best_selection(self):
        keyword_list = self.get_keyword_list()
        result = {}
        for keyword, node_id in zip(keyword_list, self.best_selection):
            result[keyword] = node_id
        return result

    def chosen(self, history_chosen, score, current_chosen_keyword_index):
        if current_chosen_keyword_index >= len(self.get_keyword_list()):
            if self.best_score < score:
                self.best_score = score
                self.best_selection = history_chosen
            else:
                return
        else:
            current_keyword = self.get_keyword_list()[current_chosen_keyword_index]

            candidate_ids = self.keyword2candidate_map[current_keyword]

            for candidate_entity_id in candidate_ids:
                extra_score = self.G.nodes[candidate_entity_id]["score"]
                for history_chosen_entity_id in history_chosen:
                    extra_score += self.G[candidate_entity_id][history_chosen_entity_id]["score"]

                self.chosen(history_chosen + [candidate_entity_id], score=score + extra_score,
                            current_chosen_keyword_index=current_chosen_keyword_index + 1)


class GlobalEntryPointLinker(EntryPointLinker):
    DEFAULT_CANDIDATE_NUM_FOR_EVERY_POINT = 3

    def __init__(self, preprocessor: Preprocessor, doc_sim_model: DocumentSimModel, kg_name_searcher: KGNameSearcher,
                 graph2vecModel: Word2VecKeyedVectors,
                 ):
        self.preprocessor = preprocessor
        self.doc_sim_model = doc_sim_model
        self.graph2vecModel = graph2vecModel
        self.kg_name_searcher = kg_name_searcher

    def link(self, query, candidate_num=DEFAULT_CANDIDATE_NUM_FOR_EVERY_POINT):
        selector = BestSeqSelector()

        keywords = self.preprocessor.extract_words_for_query(query=query)

        keyword_2_candidate_ids_map = {}
        doc_id_2_doc_retrieval_result_map = {}
        for keyword in keywords:
            valid_id_set = self.kg_name_searcher.search_by_keyword(keyword)
            doc_retrieval_result_list = self.doc_sim_model.search(query=query, top_num=candidate_num,
                                                                  valid_doc_id_set=valid_id_set)

            if len(doc_retrieval_result_list) > 0:
                keyword_2_candidate_ids_map[keyword] = doc_retrieval_result_list
            for doc_retrieval_result in doc_retrieval_result_list:
                # print("candidate keyword=%s %r" % (keyword, doc_retrieval_result))

                selector.add_candidate_for_keyword(keyword=keyword,
                                                   candidate_node_id=doc_retrieval_result.doc_id,
                                                   score=doc_retrieval_result.score)
                doc_id_2_doc_retrieval_result_map[doc_retrieval_result.doc_id] = doc_retrieval_result
        candidate_ids = selector.get_candidate_ids()

        vector_map = {}

        all_vectors = []

        for candidate_id in candidate_ids:
            vector_map[candidate_id] = self.graph2vecModel.wv[str(candidate_id)]
            all_vectors.append(vector_map[candidate_id])

        for start_id in candidate_ids:
            start_vector = vector_map[start_id]
            sim_vector = self.graph2vecModel.wv.cosine_similarities(start_vector, all_vectors)

            for end_id, score in zip(candidate_ids, sim_vector):
                selector.add_pair_score(start_id=start_id, end_id=end_id, score=score)
                # print("pair score =%r %r %r" % (start_id, end_id, score))

        best_linking_result = selector.search_best_combination()

        result = {}
        for keyword, node_id in best_linking_result.items():
            doc_retrieval_result = doc_id_2_doc_retrieval_result_map[node_id]

            result[keyword] = EntityLinkResult(mention=keyword, node_id=node_id,
                                               node_name=doc_retrieval_result.doc_name,
                                               score=doc_retrieval_result.score)

        return result



from unittest import TestCase

from sekg.ir.preprocessor.spacy import SpacyTextPreprocessor
from sekg.graph.util.name_searcher import KGNameSearcher


class TestSpacyTextPreprocessor(TestCase):
    def test_extract_keyword(self):
        name_searcher = KGNameSearcher.load(
            "E:\PycharmProjects\KGFeatureLocation\output\graph\jedite-4.3\\NameSearcher\jedite-4.3.v4.1.namesearcher"
        )
        preprocessor = SpacyTextPreprocessor(kg_name_searcher=name_searcher)
        query_str = [
            "folding bug, text is in a black hole",
            "Size of file open/save dialogs incorrect on dual displays",
            "Incorrect Python code indentation",
            "Search expression",
            "Autosave turned off for Untitled Documents",
            "Keyboard navigation through menubar inserts char into buffer",
            "Add filter to recent files menu",
            "File Open Dialog: can't handle full pathnames",
            "Provide a way of reloading changed buffers without prompting",
            "Rectangular Selection and Dragging",
            "Filter Bug in File System Browser",
            "Auto-indent breaks if previous line is a single line indent",
            "set \"Search for:\" to \"||\" check \"Regular expressions\" click \"Replace All\" version : jEdit 4.3pre8",
            "Wrong conversion to UTF-16 (with BOM)",
            "caret/display position w/ multiple views open",
        ]
        for query in query_str:
            preprocessor.extract_concept_and_operation(query)

    def test_extract_words_for_query(self):
        preprocessor = SpacyTextPreprocessor()
        test_case_list = [
            (
                "folding bug, text is in a black hole."
                , ['bug',
                   'text',
                   'black hole']
            ),

            (
                "how to get File's MD5 checksum",
                ['File', 'md5', 'checksum']

            )

        ]

        for old_str, new_str in test_case_list:
            team = preprocessor.extract_words_for_query(old_str)
            self.assertEqual(team, new_str)

    def test_clean(self):

        test_case_list = [
            (
                "This is a ArrayList, it contains all of the classes for creating user interfaces and for painting graphics and images."
                ,
                ['arraylist',
                 'contain',
                 'class',
                 'create',
                 'user',
                 'interface',
                 'painting',
                 'graphic',
                 'image']
            ),

            ("how to get File's MD5 checksum",
             ["file", "md5", "checksum"])
        ]
        preprocessor = SpacyTextPreprocessor()

        for old_str, keywords in test_case_list:
            self.assertEqual(keywords, preprocessor.clean(old_str))



from unittest import TestCase

from sekg.metric.ir import InformationRetrievalMetric


class TestInformationRetrievalMetric(TestCase):
    def test_precision(self):
        result, separate_avg = InformationRetrievalMetric.precision(
            [([1, 1, 1], 10), ([0, 0], 2), ([0, 1, 1, 1, 0, 1, 0, 1, 1, 0], 10)])
        self.assertEqual(result, 0.5333333333333333)
        self.assertEqual(separate_avg, [1, 0, 0.6])

        result, separate_avg = InformationRetrievalMetric.precision(([1, 1, 1], 10))

        self.assertEqual(result, 1)
        self.assertEqual(separate_avg, [1])

    def test_recall(self):
        result, separate_avg = InformationRetrievalMetric.recall(
            [([1, 1, 1], 10), ([0, 0], 2), ([0, 1, 1, 1, 0, 1, 0, 1, 1, 0], 10)])

        self.assertEqual(result, 0.3)
        self.assertEqual(separate_avg, [0.3, 0, 0.6])

        result, separate_avg = InformationRetrievalMetric.recall(([1, 1, 1], 10))

        self.assertEqual(result, 0.3)
        self.assertEqual(separate_avg, [0.3])

    def test_mrr(self):
        result, separate_avg = InformationRetrievalMetric.mrr(
            [([1, 1, 1], 10), ([0, 0], 2), ([0, 1, 1, 1, 0, 1, 0, 1, 1, 0], 10)])

        self.assertEqual([1, 0, 0.5], separate_avg)
        self.assertEqual(0.5, result)

        result, separate_avg = InformationRetrievalMetric.mrr(([1, 1, 1], 10))

        self.assertEqual(1, result)
        self.assertEqual([1], separate_avg)

    def test_map(self):
        result, separate_avg = InformationRetrievalMetric.map(
            [([1, 1, 1], 10), ([0, 0], 2), ([0, 1, 1, 1, 0, 1, 0, 1, 1, 0], 10)])

        self.assertEqual([0.3, 0, 0.38749999999999996], separate_avg)
        self.assertEqual(0.22916666666666666, result)

        result, separate_avg = InformationRetrievalMetric.map(([1, 1, 1], 10))

        self.assertEqual(0.3, result)
        self.assertEqual([0.3], separate_avg)
        result, separate_avg = InformationRetrievalMetric.map(([1, 0, 1, 0, 0, 1, 0, 0], 3))

        self.assertEqual(0.7222222222222222, result)
        self.assertEqual([0.7222222222222222], separate_avg)

    def test_f1(self):
        result, separate_avg = InformationRetrievalMetric.f1(
            [([1, 1, 1], 10), ([0, 0], 2), ([0, 1, 1, 1, 0, 1, 0, 1, 1, 0], 10)])

        self.assertEqual([0.4615384615384615, 0.0, 0.6], separate_avg)
        self.assertEqual(0.35384615384615387, result)

        result, separate_avg = InformationRetrievalMetric.f1(([1, 1, 1], 10))

        self.assertEqual(0.4615384615384615, result)
        self.assertEqual([0.4615384615384615], separate_avg)
        result, separate_avg = InformationRetrievalMetric.f1(([1, 0, 1, 0, 0, 1, 0, 0], 3))

        self.assertEqual(0.5454545454545454, result)
        self.assertEqual([0.5454545454545454], separate_avg)



#!/usr/bin/python
# -*- coding: UTF-8 -*-
import os

import gensim
import numpy as np

from sekg.ir.doc.wrapper import PreprocessMultiFieldDocumentCollection
from sekg.ir.models.base import DocumentSimModel
from sekg.util.file import DirUtil
from sekg.util.vector_util import VectorUtil


class TFIDFW2VFLModel(DocumentSimModel):
    __tfidf_model_name__ = "tfidf.model"
    __corpus_name__ = 'corpus.mm'
    __w2v_model_name__ = "word2vec.model"
    __tfidf_w2v_model_name__ = "word2vec.tfidf.model"
    __dictionary_name__ = 'dictionary.dict'
    __sim_index__ = "tfidfw2v.sim.index"

    __entity_collection_name__ = "doc.pre.collection"
    DEFAULT_EMBEDDING_SIZE = 100
    DEFAULT_TFIDF_KEYWORD_NUM = 20

    def __init__(self, name, model_dir_path, **config):

        """
        init the lsi model with
        :param model_dir_path:
        """
        super().__init__(name, model_dir_path, **config)

        self.tfidf_model = None
        self.tfidf_model_path = None
        self.w2v_model = None
        self.tfidf_w2v_model = None
        self.corpus = None
        self.dict = None
        self.similarity_index = None
        self.__init_sub_model_path__()
        self.preprocessor = None
        self.preprocess_doc_collection = None

        self.NP_VECTOR_NOT_EXIST = None
        self.__init_embedding_size(TFIDFW2VFLModel.DEFAULT_EMBEDDING_SIZE)
        self.entity_collection_path = None

    def __init_sub_model_path__(self):
        if self.model_dir_path is None:
            return
            # init the paths of models
        self.corpus_path = os.path.join(self.model_dir_path, self.__corpus_name__)
        self.dictionary_path = os.path.join(self.model_dir_path, self.__dictionary_name__)
        if self.tfidf_model_path is None:
            self.tfidf_model_path = os.path.join(self.model_dir_path, self.__tfidf_model_name__)

        self.w2v_model_path = os.path.join(self.model_dir_path, self.__w2v_model_name__)
        self.tfidf_w2v_model_path = os.path.join(self.model_dir_path, self.__tfidf_w2v_model_name__)

        self.entity_collection_path = os.path.join(self.model_dir_path, self.__entity_collection_name__)

        self.sim_index_dir = os.path.join(self.model_dir_path, "index")

        self.sim_index_path = os.path.join(self.sim_index_dir, self.__sim_index__)

        DirUtil.create_file_dir(self.model_dir_path)
        DirUtil.create_file_dir(self.sim_index_dir)

    def __init_document_collection(self, preprocess_doc_collection):
        self.preprocess_doc_collection = preprocess_doc_collection
        self.preprocessor = preprocess_doc_collection.get_preprocessor()
        self.field_set = preprocess_doc_collection.get_field_set()

    def init_model_as_submodel(self):
        """
        init the model
        :return:
        """
        print("init_sub_model_path")
        self.__init_sub_model_path__()

        print("loading doc_collection")
        preprocess_doc_collection = PreprocessMultiFieldDocumentCollection.load(self.entity_collection_path)
        self.preprocessor = preprocess_doc_collection.get_preprocessor()
        self.field_set = preprocess_doc_collection.get_field_set()

        print("loading the tfidf model models")
        self.tfidf_model = gensim.models.TfidfModel.load(self.tfidf_model_path)
        print("loading the Word2vec models")
        self.w2v_model = gensim.models.keyedvectors.Word2VecKeyedVectors.load(self.w2v_model_path)
        print("loading the tfidf w2v models")
        self.tfidf_w2v_model = gensim.models.keyedvectors.Word2VecKeyedVectors.load(self.tfidf_w2v_model_path)

        self.dict = gensim.corpora.Dictionary.load(self.dictionary_path)
        print("All document number: ", self.dict.num_docs)
        print("All words number: ", self.dict.num_pos)
        print("load models done")

    def init_model(self):
        """
        init the model
        :return:
        """
        print("init_sub_model_path")
        self.__init_sub_model_path__()

        print("loading doc_collection")
        preprocess_doc_collection = PreprocessMultiFieldDocumentCollection.load(self.entity_collection_path)
        self.__init_document_collection(preprocess_doc_collection)

        print("loading the tfidf model models")
        self.tfidf_model = gensim.models.TfidfModel.load(self.tfidf_model_path)
        print("loading the Word2vec models")
        self.w2v_model = gensim.models.keyedvectors.Word2VecKeyedVectors.load(self.w2v_model_path)
        print("loading the tfidf w2v models")
        self.tfidf_w2v_model = gensim.models.keyedvectors.Word2VecKeyedVectors.load(self.tfidf_w2v_model_path)

        self.corpus = gensim.corpora.MmCorpus(self.corpus_path)
        self.dict = gensim.corpora.Dictionary.load(self.dictionary_path)
        print("All document number: ", self.dict.num_docs)
        print("All words number: ", self.dict.num_pos)
        print("load models done")

    def __init_embedding_size(self, embedding_size):
        self.embedding_size = embedding_size
        self.NP_VECTOR_NOT_EXIST = np.zeros(embedding_size)
        self.NP_VECTOR_NOT_EXIST[0] = 1e-07

    def get_full_doc_score_vec(self, query):
        full_entity_score_vec = self.get_cache_score_vector(query)
        if full_entity_score_vec is not None:
            return full_entity_score_vec

        query_words = self.preprocessor.clean(query)
        query_vec = self.get_tfidf_w2v_vec(query_words)
        full_entity_score_vec = self.tfidf_w2v_model.similar_by_vector(query_vec, topn=None)
        full_entity_score_vec = (full_entity_score_vec + 1) / 2
        # save the result to cache
        self.cache_entity_score_vector(query, full_entity_score_vec)
        return full_entity_score_vec

    def train_from_document_collection(self, preprocess_document_collection, embedding_size=100,
                                       pretrain_w2v_path=None, pretrain_tfidf_path=None, pretrain_dictionary_path=None):
        print("start training")
        self.__init_embedding_size(embedding_size)
        self.__init_document_collection(preprocess_document_collection)

        corpus_clean_text = []
        preprocess_multi_field_doc_list = preprocess_document_collection.get_all_preprocess_document_list()
        for docno, multi_field_doc in enumerate(preprocess_multi_field_doc_list):
            corpus_clean_text.append(multi_field_doc.get_document_text_words())

        print("corpus len=%d" % len(corpus_clean_text))

        if pretrain_dictionary_path is not None:
            print("pretrain pretrain dictionary path is given, loading")
            self.dict = gensim.corpora.Dictionary.load(pretrain_dictionary_path)
        else:
            print("Dictionary init...")
            self.dict = gensim.corpora.Dictionary(corpus_clean_text)
            print("Dictionary init complete")

        print("parse to bow corpus")
        self.corpus = [self.dict.doc2bow(text) for text in corpus_clean_text]
        print("parse to bow corpus complete")

        if pretrain_w2v_path is not None:
            print("pretrain Word2vec path is given, loading")
            self.w2v_model = gensim.models.keyedvectors.Word2VecKeyedVectors.load(pretrain_w2v_path)
        else:
            print("pretrain Word2vec path is not given, loading")

            print("Word2Vec Training...")
            w2v_model = gensim.models.Word2Vec(sentences=corpus_clean_text, size=embedding_size, min_count=1)
            print("Word2Vec Train complete")
            self.w2v_model = w2v_model.wv

        if pretrain_tfidf_path is not None:
            print("pretrain tfidf path is given, loading")
            self.tfidf_model = gensim.models.TfidfModel.load(pretrain_tfidf_path)

        else:
            print("tfidf Training...")
            self.tfidf_model = gensim.models.TfidfModel(corpus=self.corpus, id2word=self.dict)
            print("tfidf Train complete")

        self.tfidf_w2v_model = self.train_tfidf_weight_w2v_model(corpus_clean_text, embedding_size)

    def train_tfidf_weight_w2v_model(self, corpus_clean_text, embedding_size):
        self.__init_embedding_size(embedding_size=embedding_size)

        tfidf_w2v_model = gensim.models.keyedvectors.Word2VecKeyedVectors(vector_size=embedding_size)
        tfidf_vector_list = []
        tfidf_vector_index_str_list = []
        for index, doc_clean_text in enumerate(corpus_clean_text):
            vector = self.get_tfidf_w2v_vec(doc_clean_text)
            tfidf_vector_list.append(vector)
            tfidf_vector_index_str_list.append(str(index))
        tfidf_w2v_model.add(entities=tfidf_vector_index_str_list, weights=tfidf_vector_list, replace=True)

        return tfidf_w2v_model

    def get_tfidf_w2v_vec(self, word_str_list):
        """
        get the average word2vec for list of str
        :param word_str_list:  list of str
        :return: np.array()
        """
        doc_word_vectors = []
        doc_bow = self.dict.doc2bow(word_str_list)
        doc_tfidf = self.tfidf_model[doc_bow]
        sorted_doc_tfidf = sorted(doc_tfidf, key=lambda x: x[1], reverse=True)
        sorted_doc_tfidf = sorted_doc_tfidf[:self.DEFAULT_TFIDF_KEYWORD_NUM]

        weight = []
        for word_id, tf_tfidf in sorted_doc_tfidf:
            word = self.dict[word_id]
            if word not in self.w2v_model.vocab:
                continue
            doc_word_vectors.append(self.w2v_model[word])
            weight.append(tf_tfidf)
        if len(doc_word_vectors) == 0:
            return self.NP_VECTOR_NOT_EXIST
        if len(weight) == 0:
            return self.NP_VECTOR_NOT_EXIST
        try:
            avg_vector = VectorUtil.get_weight_mean_vec(doc_word_vectors, weight)
            return avg_vector
        except Exception as e:
            print(e)
            # print(doc_word_vectors)
            # print(weight)
            return self.NP_VECTOR_NOT_EXIST

    def train_from_doc_collection_with_preprocessor(self, doc_collection, **config):
        embedding_size = self.DEFAULT_EMBEDDING_SIZE
        pretrain_w2v_path = None
        pretrain_tfidf_path = None
        pretrain_dictionary_path = None

        if "embedding_size" in config.keys():
            embedding_size = config["embedding_size"]
        if "pretrain_w2v_path" in config.keys():
            pretrain_w2v_path = config["pretrain_w2v_path"]
        if "pretrain_tfidf_path" in config.keys():
            pretrain_tfidf_path = config["pretrain_tfidf_path"]
        if "pretrain_dictionary_path" in config.keys():
            pretrain_dictionary_path = config["pretrain_dictionary_path"]

        if isinstance(doc_collection, PreprocessMultiFieldDocumentCollection):
            self.train_from_document_collection(preprocess_document_collection=doc_collection,
                                                embedding_size=embedding_size,
                                                pretrain_w2v_path=pretrain_w2v_path,
                                                pretrain_tfidf_path=pretrain_tfidf_path,
                                                pretrain_dictionary_path=pretrain_dictionary_path
                                                )

    def save(self, model_dir_path):
        """
        save model to the model_dir_path
        :param model_dir_path: the dir to save the model
        :return:
        """
        super().save(model_dir_path)

        self.__init_sub_model_path__()

        self.dict.save(self.dictionary_path)

        print("build dictionary in %s" % self.dictionary_path)

        gensim.corpora.MmCorpus.serialize(self.corpus_path, self.corpus)
        print("save the corpus to %s" % self.corpus_path)

        self.w2v_model.save(self.w2v_model_path)
        print("Word2vec save finish , save to %s" % self.w2v_model_path)

        self.tfidf_model.save(self.tfidf_model_path)
        print("tfidf save finish , save to %s" % self.tfidf_model_path)

        self.tfidf_w2v_model.save(self.tfidf_w2v_model_path)
        print("TF IDF Word2vec Training finish , save to %s" % self.tfidf_w2v_model_path)

        print("entity collection saving...")
        self.preprocess_doc_collection.save(self.entity_collection_path)
        print(
            "entity collection finish saving , save to %s, %r" % (
                self.entity_collection_path, self.preprocess_doc_collection))



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from kgtools.annotation import Lazy


class Synset:
    def __init__(self, terms):
        self.terms = terms
        self.text = "<%s>" % ", ".join([term for term in list(sorted(self.terms, key=lambda x: x))])
        self.term2count = {}

    @Lazy
    def key(self):
        count2terms = {}
        for term in self.terms:
            count = self.term2count.get(term, 0)
            if count not in count2terms:
                count2terms[count] = set()
            count2terms[count].add(term)

        _, candidates = max(count2terms.items(), key=lambda item: item[0])
        return max(candidates, key=lambda x: len(x))

    def init_count(self, term_count):
        for term in self.terms:
            self.term2count[term] = term_count.get(term, 0)

    def __iadd__(self, other):
        self.terms |= other.terms
        self.term2count.update(other.term2count)
        return self

    def __add__(self, other):
        synset = Synset(self.terms | other.terms)
        synset.term2count = dict(set(self.term2count.items()) | set(other.term2count.items()))
        return synset

    def __str__(self):
        return self.text

    def __hash__(self):
        return hash(str(self))

    def __iter__(self):
        return iter(self.terms)

    def __eq__(self, other):
        return hash(self) == hash(other)



import random

import gensim

from sekg.ir.preprocessor.base import Preprocessor


class MultiFieldDocumentCollection(gensim.utils.SaveLoad):
    """
    This class is a wrapper for multi field document collection. It contain many MultiFieldDocument instance. Each one stand for a document.
    """

    def __init__(self, documents=None):
        self.documents = []
        self.field_set = set([])
        if documents:
            for document in documents:
                self.add_document(document)

        self.doc_id_2_documents_map = {}
        self.doc_id_2_doc_index_map = {}

    def get_num(self):
        return len(self.documents)

    def add_document(self, document):
        """
        :param document:
        :return: False, the doc with the id already exist. True, add new doc success
        """
        doc_id = document.id
        if doc_id in self.doc_id_2_documents_map.keys():
            return False
        self.documents.append(document)
        self.field_set.update(document.get_field_set())
        self.doc_id_2_documents_map[doc_id] = document
        self.doc_id_2_doc_index_map[doc_id] = len(self.documents) - 1
        return True

    def get_by_id(self, id):
        return self.doc_id_2_documents_map.get(id, None)

    def get_by_index(self, index):
        index = int(index)
        if index < 0 or index >= len(self.documents):
            return None
        return self.documents[index]

    def __str__(self):
        return "Documents(Num=%d)" % (self.get_num())

    def clear(self):
        self.documents = []
        self.field_set = set([])

    def get_field_set(self):
        return self.field_set

    def get_document_list(self):
        return self.documents

    def add_field_to_doc(self, doc_id, field_name, value):
        doc = self.get_by_id(id=doc_id)
        if doc is None:
            return
        doc.add_field(field_name, value)

    def doc_index_to_doc_id(self, index):
        return self.get_by_index(index).get_document_id()

    def doc_id_to_doc_index(self, doc_id):
        return self.doc_id_2_doc_index_map.get(doc_id, None)

    def get_doc_id_2_doc_index_map(self):
        return self.doc_id_2_doc_index_map

    def exist(self, id):
        """
        check if the doc with the specify id exist
        :param id: the doc id
        :return: True, exist,False, not exist
        """
        doc = self.get_by_id(id)
        if doc == None:
            return False
        return True

    def size(self):
        """
        get the doc num in this collection
        :return: the doc size
        """
        return self.get_num()

    def pretty_print_by_id(self, id):
        if not self.exist(id):
            print("Not exist doc for id=%r" % id)
        raw_doc = self.get_by_id(id)
        raw_doc.pretty_print()

    def random_docs(self, random_num):
        num = self.get_num()
        random_index_list = list(range(0, num))
        random.shuffle(random_index_list)
        random_index_list = random_index_list[:random_num]

        docs = []
        for index in random_index_list:
            doc = self.get_by_index(index)
            docs.append(doc)
        return docs

    def random_doc(self, ):
        num = self.get_num()
        random_index = random.randint(0, num)
        doc = self.get_by_index(random_index)
        return doc

    def doc_id_set_2_doc_index_set(self, doc_id_set):
        doc_index_set = set([])
        for doc_id in doc_id_set:
            doc_index = self.doc_id_to_doc_index(doc_id)
            if doc_index is not None:
                doc_index_set.add(doc_index)
        return doc_index_set

    def doc_index_set_2_doc_id_set(self, doc_index_set):
        doc_id_set = set([])
        for doc_index in doc_index_set:
            doc_id = self.doc_index_to_doc_id(doc_index)
            if doc_id is not None:
                doc_id_set.add(doc_id)
        return doc_id_set

    def sub_document_collection(self, doc_id_set):
        collection = MultiFieldDocumentCollection()
        for doc_id in doc_id_set:
            doc = self.get_by_id(doc_id)
            if doc != None:
                collection.add_document(doc)
        return collection


## todo: add iterate function and set function
class PreprocessMultiFieldDocumentCollection(gensim.utils.SaveLoad):
    """
        This class is a wrapper for multi field document collection. It contain many MultiFieldDocument instance. Each one stand for a document.
        But it contain preproccessed doc and the original doc.
    """

    def __init__(self, preprocessor: Preprocessor):
        self.raw_doc_collection = MultiFieldDocumentCollection()
        self.preprocess_doc_collection = MultiFieldDocumentCollection()

        if preprocessor is None:
            raise Exception("preprocessor is None")

        if isinstance(preprocessor, Preprocessor):
            self.preprocessor = preprocessor
        else:
            raise Exception("preprocessor is not a Preprocessor instance")

    @staticmethod
    def create_from_doc_collection(preprocessor: Preprocessor, doc_collection: MultiFieldDocumentCollection):
        """
        create the preprocessed document collection
        :param preprocessor: preprocessor use to preprocess the document
        :param doc_collection: the document collection need to preprocess
        :return: PreprocessMultiFieldDocumentCollection
        """
        pre_doc_collection = PreprocessMultiFieldDocumentCollection(preprocessor)
        pre_doc_collection.add_raw_doc_collection(doc_collection)
        return pre_doc_collection

    def get_field_set(self):
        return self.raw_doc_collection.get_field_set()

    def get_preprocessor(self):
        return self.preprocessor

    def get_num(self):
        return self.raw_doc_collection.get_num()

    def add_raw_doc_and_preprocess_doc_pair(self, raw_doc, preprocess_doc):
        self.raw_doc_collection.add_document(raw_doc)
        self.preprocess_doc_collection.add_document(preprocess_doc)

    def get_raw_doc_and_preprocess_doc_pair(self, index):
        return (
            self.get_raw_doc_by_index(index),
            self.get_preprocess_doc_by_index(index)
        )

    def get_doc_pair_by_index(self, index):
        return (self.get_raw_doc_by_index(index), self.get_preprocess_doc_by_index(index))

    def get_doc_pair_by_id(self, id):
        return (self.get_raw_doc_by_id(id), self.get_preprocess_doc_by_id(id))

    def get_raw_doc_by_index(self, index):
        """
        get the
        :param index:
        :return:
        """
        return self.raw_doc_collection.get_by_index(index)

    def get_raw_doc_by_id(self, id):
        return self.raw_doc_collection.get_by_id(id)

    def get_preprocess_doc_by_id(self, id):
        return self.preprocess_doc_collection.get_by_id(id)

    def get_preprocess_doc_by_index(self, index):
        """
        the preprocess doc is list of str
        :param index:
        :return: a list of str
        """
        return self.preprocess_doc_collection.get_by_index(index)

    def __str__(self):
        return "Documents(Num=%d)" % (self.get_num())

    def add_raw_doc(self, raw_doc):
        if isinstance(raw_doc, MultiFieldDocument):
            self.raw_doc_collection.add_document(raw_doc)

            preprocess_doc = MultiFieldDocument(id=raw_doc.get_document_id(), name=raw_doc.get_name())
            for field, field_doc in raw_doc.get_all_field_doc_map().items():
                preprocess_field_doc_words = self.preprocessor.clean(field_doc)
                preprocess_doc.add_field(field_name=field, field_document=" ".join(preprocess_field_doc_words))
            self.preprocess_doc_collection.add_document(preprocess_doc)

    def add_raw_doc_collection(self, raw_doc_collection):
        for index, raw_doc in enumerate(raw_doc_collection.get_document_list()):
            self.add_raw_doc(raw_doc)

    def clear(self):
        self.raw_doc_collection.clear()
        self.preprocess_doc_collection.clear()

    def get_all_preprocess_documents_text(self):
        """
        get list of str
        :return:
        """
        result = []
        for preprocess_doc in self.preprocess_doc_collection.documents:
            result.append(preprocess_doc.get_document_text())
        return result

    def get_all_preprocess_documents_text_words(self):
        """
        get list of str
        :return:
        """
        result = []
        for preprocess_doc in self.preprocess_doc_collection.documents:
            result.append(preprocess_doc.get_document_text_words())
        return result

    def get_all_preprocess_document_list(self):
        """
        get the list of all preprocess_document( containing multifield)
        :return:
        """
        return self.preprocess_doc_collection.get_document_list()

    def get_all_preprocess_field_words(self, field_name):
        """
        get list of str
        :return:
        """
        result = []
        for preprocess_doc in self.preprocess_doc_collection.documents:
            result.append(preprocess_doc.get_doc_words_by_field(field_name))
        return result

    def get_by_index(self, index):
        return self.raw_doc_collection.get_by_index(index)

    def doc_index_to_doc_id(self, index):
        return self.raw_doc_collection.doc_index_to_doc_id(index)

    def doc_id_to_doc_index(self, doc_id):
        return self.raw_doc_collection.doc_id_to_doc_index(doc_id)

    def __repr__(self):
        return "<PreprocessMultiFieldDocumentCollection num=%r>" % (self.raw_doc_collection.get_num())

    def exist(self, id):
        """
        check if the doc with the specify id exist
        :param id: the doc id
        :return: True, exist,False, not exist
        """
        doc = self.get_raw_doc_by_id(id)
        if doc == None:
            return False
        return True

    def size(self):
        """
        get the doc num in this collection
        :return: the doc size
        """
        return self.get_num()

    def pretty_print_by_id(self, id):
        if not self.exist(id):
            print("Not exist doc for id=%r" % id)
        (raw_doc, preprocess_doc) = self.get_doc_pair_by_id(id)
        print("raw doc:")
        raw_doc.pretty_print()
        print("preprocessed doc:")
        preprocess_doc.pretty_print()

    def random_docs(self, random_num):
        return self.preprocess_doc_collection.random_docs(random_num)

    def random_doc(self):
        return self.preprocess_doc_collection.random_doc()

    def get_doc_id_2_doc_index_map(self):
        return self.preprocess_doc_collection.get_doc_id_2_doc_index_map()

    def doc_id_set_2_doc_index_set(self, doc_id_set):
        return self.preprocess_doc_collection.doc_id_set_2_doc_index_set(doc_id_set)

    def doc_index_set_2_doc_id_set(self, doc_index_set):

        return self.preprocess_doc_collection.doc_index_set_2_doc_id_set(doc_index_set)

    def sub_document_collection(self, doc_id_set):

        collection = PreprocessMultiFieldDocumentCollection(self.preprocessor)

        for doc_id in doc_id_set:
            raw_doc, pre_doc = self.get_doc_pair_by_id(doc_id)
            if raw_doc == None or pre_doc == None:
                continue

            collection.add_raw_doc_and_preprocess_doc_pair(raw_doc, pre_doc)
        return collection


## todo: try to reduce the MultiFieldDocument and PreprocessorMultiFieldDocument

class MultiFieldDocument:
    """
    This class is a wrapper for the document with multi field.
    For example, for a wikipedia doc. we can has two field "title","body".
    for a stack overflow question, we could have "question title","tags","question body","accept answer".
    For this purpose, this class is used to wrapper the doc like this has multi field.

    """

    def __init__(self, id, name, field_doc=None):
        if field_doc is None:
            field_doc = {}
        self.id = id
        self.name = name
        self.field_doc = field_doc

    def add_field(self, field_name, field_document):
        self.field_doc[field_name] = field_document

    def get_field_set(self):
        return self.field_doc.keys()

    def get_doc_text_by_field(self, field_name):
        """

        :param field_name:
        :return:
        """
        if field_name in self.field_doc.keys():
            return self.field_doc[field_name]
        else:
            return ""

    def get_doc_words_by_field(self, field_name):
        """
        the return result could be list of str or a text
        :param field_name:
        :return:
        """
        text = self.get_doc_text_by_field(field_name)
        return text.split()

    def get_all_field_doc_map(self):
        return self.field_doc

    def get_name(self):
        return self.name

    def get_document_id(self):
        return self.id

    def get_document_text(self):
        """
        get all the text from this MultiFieldDocument by conbining text from all field.
        :return: return a str
        """
        docs = []
        for field_name in self.get_field_set():
            doc = self.get_doc_text_by_field(field_name=field_name)
            docs.append(doc)
        return "\n".join(docs)

    def get_document_text_words(self):
        """
        get all the text from this MultiFieldDocument by conbining text from all field.
        :return: return a iteration of str. each one is a word in doc field.
        """

        docs = []
        for field_name in self.get_field_set():
            doc = self.get_doc_words_by_field(field_name=field_name)
            docs.extend(doc)
        return docs

    def pretty_print(self):
        print("doc id=%r" % self.id)
        field_doc = self.get_all_field_doc_map()
        for field, doc in field_doc.items():
            print("field=%r doc=%r" % (field, doc))
        print("-" * 20)

    def __repr__(self):
        return "<MultiFieldDocument= id=%d name=%r doc=%r>" % (self.get_document_id(), self.get_name(), self.field_doc)



#!/usr/bin/python
# -*- coding: UTF-8 -*-
import pickle
# from functools import partial
# from multiprocessing import Pool
from pathlib import Path

import gensim
import math
import numpy as np
# from gensim.summarization.bm25 import BM25
# from gensim.utils import effective_n_jobs
from six import iteritems
from sekg.ir.doc.wrapper import PreprocessMultiFieldDocumentCollection
from sekg.ir.models.base import DocumentSimModel
from sekg.util.annotation import catch_exception


# todo: speed up
class BM25Model(DocumentSimModel):
    """
    This class is used the BM25 model to compute the document similarity score.
    The implementation from BM25 is from gensim.
    """
    __bm25_model_name__ = "bm25.model"
    __doc_collection_name__ = "doc.pre.collection"
    __dictionary_name__ = "corpus.dict"

    def __init__(self, name, model_dir_path, **config):

        """
        init the model with model_dir_path:
        """
        super().__init__(name, model_dir_path, **config)

        self.bm25_model = None
        self.corpus = None
        self.dict = None
        self.__init_sub_model_path__()

        self.preprocessor = None
        self.preprocess_doc_collection = None

    def __init_sub_model_path__(self):
        if self.model_dir_path is None:
            return
        model_dir = Path(self.model_dir_path)
        self.bm25_model_path = str(model_dir / self.__bm25_model_name__)
        self.entity_collection_path = str(model_dir / self.__doc_collection_name__)
        self.dictionary_path = str(model_dir / self.__dictionary_name__)
        model_dir.mkdir(parents=True, exist_ok=True)

    def init_model_as_submodel(self):
        """
        init the model
        :return:
        """
        self.init_model()

    @catch_exception
    def init_model(self):
        """
        init the model
        :return:
        """
        self.__init_sub_model_path__()
        print("loading the bm25 models")
        with open(self.bm25_model_path, 'rb') as fr:
            self.bm25_model = pickle.load(fr)

        self.dict = gensim.corpora.Dictionary.load(self.dictionary_path)
        print("All document number: ", self.dict.num_docs)
        print("All words number: ", self.dict.num_pos)

        preprocess_doc_collection = PreprocessMultiFieldDocumentCollection.load(self.entity_collection_path)
        self.__init_document_collection(preprocess_doc_collection)

    def __init_document_collection(self, preprocess_doc_collection: PreprocessMultiFieldDocumentCollection):
        self.set_preprocess_doc_collection(preprocess_doc_collection)
        self.set_preprocessor(preprocess_doc_collection.get_preprocessor())

    def train_from_doc_collection_with_preprocessor(self, doc_collection: PreprocessMultiFieldDocumentCollection,
                                                    **config):
        print("start training")
        self.__init_document_collection(doc_collection)
        corpus_clean_text = []
        preprocess_multi_field_doc_list = doc_collection.get_all_preprocess_document_list()

        for docno, multi_field_doc in enumerate(preprocess_multi_field_doc_list):
            corpus_clean_text.append(multi_field_doc.get_document_text_words())
        print("corpus len=%d" % len(corpus_clean_text))

        self.dict = gensim.corpora.Dictionary(corpus_clean_text)
        print("Dictionary init complete")
        self.corpus = [self.dict.doc2bow(text) for text in corpus_clean_text]

        print("bm25 Training...")
        self.bm25_model = BM25(corpus=self.corpus)
        print("bm25 compelete...")

    @catch_exception
    def save(self, model_dir_path):
        """
        save model to the model_dir_path
        :param model_dir_path: the dir to save the model
        :return:
        """
        super().save(model_dir_path)
        self.__init_sub_model_path__()

        self.dict.save(self.dictionary_path)
        print("build dictionary in %s" % self.dictionary_path)

        print("save bm25 model into ", self.bm25_model_path)
        with open(self.bm25_model_path, 'wb') as fw:
            pickle.dump(self.bm25_model, fw)

        print("entity collection saving...")
        self.preprocess_doc_collection.save(self.entity_collection_path)
        print(
            "entity collection finish saving , save to %s, %r" % (
                self.entity_collection_path, self.preprocess_doc_collection))

    def get_full_doc_score_vec(self, query):
        """
        score vector is a vector v=[0.5,2.0,3.0], v[0] means that the document 'd'
         whose index is 0 in DocumentCollection, its score with query is 0.5.
        :param query: a str stands for the query.
        :return: get all document similar score with query as a numpy vector.
        """

        full_entity_score_vec = self.get_cache_score_vector(query)
        if full_entity_score_vec is not None:
            return full_entity_score_vec

        query_words = self.preprocessor.clean(query)
        query_bow = self.dict.doc2bow(query_words)
        full_entity_score_vec = np.array(self.bm25_model.get_scores(query_bow))
        self.cache_entity_score_vector(query, full_entity_score_vec)
        return full_entity_score_vec

    def similarity_4_doc_id_pair(self, doc_id1, doc_id2):
        # todo: implement in the base class or set as a abstract class
        # todo: fix this method. Change the BM25 to compute by numpy for speed up
        # todo: add checking for the doc_id1

        doc1 = self.doc_id2doc(doc_id1)
        pos2 = self.doc_id2doc_index(doc_id2)

        if doc1 is None or pos2 is None:
            return 0.0
        words = doc1.get_document_text_words()

        score = self.bm25_model.get_score(words, pos2)

        return score

    def similarity_by_doc_id(self, doc_id):
        """

        :param doc_id:
        :return:
        """
        # todo: implement in the base class or set as a abstract class
        # todo: fix this method. Change the BM25 to compute by numpy for speed up
        doc = self.doc_id2doc(doc_id)
        words = doc.get_document_text_words()
        full_entity_score_vec = np.array(self.bm25_model.get_scores(words))
        return full_entity_score_vec


# !/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Licensed under the GNU LGPL v2.1 - http://www.gnu.org/licenses/lgpl.html

"""This module contains function of computing rank scores for documents in
corpus and helper class `BM25` used in calculations. Original algorithm
descibed in [1]_, also you may check Wikipedia page [2]_.


.. [1] Robertson, Stephen; Zaragoza, Hugo (2009).  The Probabilistic Relevance Framework: BM25 and Beyond,
       http://www.staff.city.ac.uk/~sb317/papers/foundations_bm25_review.pdf
.. [2] Okapi BM25 on Wikipedia, https://en.wikipedia.org/wiki/Okapi_BM25



Examples
--------

.. sourcecode:: pycon

    # >>> from gensim.summarization.bm25 import get_bm25_weights
    # >>> corpus = [
    # ...     ["black", "cat", "white", "cat"],
    # ...     ["cat", "outer", "space"],
    # ...     ["wag", "dog"]
    # ... ]
    # >>> result = get_bm25_weights(corpus, n_jobs=-1)


Data:
-----
.. data:: PARAM_K1 - Free smoothing parameter for BM25.
.. data:: PARAM_B - Free smoothing parameter for BM25.
.. data:: EPSILON - Constant used for negative idf of document in corpus.

"""

PARAM_K1 = 1.5
PARAM_B = 0.75
EPSILON = 0.25


class BM25(object):
    """Implementation of Best Matching 25 ranking function.

    Attributes
    ----------
    corpus_size : int
        Size of corpus (number of documents).
    avgdl : float
        Average length of document in `corpus`.
    doc_freqs : list of dicts of int
        Dictionary with terms frequencies for each document in `corpus`. Words used as keys and frequencies as values.
    idf : dict
        Dictionary with inversed documents frequencies for whole `corpus`. Words used as keys and frequencies as values.
    doc_len : list of int
        List of document lengths.
    """

    def __init__(self, corpus):
        """
        Parameters
        ----------
        corpus : list of list of str
            Given corpus.

        """
        self.corpus_size = len(corpus)
        self.avgdl = 0
        self.doc_freqs = []
        self.idf = {}
        self.wf = {}
        self.doc_len = []
        # self._initialize(corpus)
        self._initialize_with_bow(corpus)
        self.word2score = {}
        self.K = (PARAM_K1 * (1 - PARAM_B + PARAM_B * np.array(self.doc_len) / self.avgdl)).tolist()  # document*1
        self.k1 = PARAM_K1 + 1

    def _initialize(self, corpus):
        """Calculates frequencies of terms in documents and in corpus. Also computes inverse document frequencies."""
        # words = list(set([y for x in corpus for y in x]))
        nd = {}  # word -> number of documents with word
        num_doc = 0
        for index, document in enumerate(corpus):
            self.doc_len.append(len(document))
            num_doc += len(document)
            # all_frequencies = dict(zip(words, [0] * len(words)))

            frequencies = {}
            for word in document:
                if word not in frequencies:
                    frequencies[word] = 0
                frequencies[word] += 1
                # all_frequencies[word] += 1
            self.doc_freqs.append(frequencies)

            for word in frequencies:
                if word not in self.wf.keys():
                    self.wf[word] = []
                self.wf[word].append([index, frequencies[word]])
                if word not in nd:
                    nd[word] = 0
                nd[word] += 1
        self.avgdl = float(num_doc) / self.corpus_size
        # collect idf sum to calculate an average idf for epsilon value
        idf_sum = 0
        # collect words with negative idf to set them a special epsilon value.
        # idf can be negative if word is contained in more than half of documents
        negative_idfs = []
        for word, freq in iteritems(nd):
            idf = math.log(self.corpus_size - freq + 0.5) - math.log(freq + 0.5)
            self.idf[word] = idf
            idf_sum += idf
            if idf < 0:
                negative_idfs.append(word)
        self.average_idf = float(idf_sum) / len(self.idf)

        eps = EPSILON * self.average_idf
        for word in negative_idfs:
            self.idf[word] = eps

    def _initialize_with_bow(self, corpus):
        """Calculates frequencies of terms in documents and in corpus. Also computes inverse document frequencies."""
        nd = {}  # word -> number of documents with word
        num_doc = 0
        for index, document in enumerate(corpus):
            self.doc_len.append(len(document))
            frequencies = {}
            for word, fre in document:
                num_doc += fre
                frequencies[word] = fre
                if word not in nd:
                    nd[word] = 0
                nd[word] += 1
                if word not in self.wf.keys():
                    self.wf[word] = []
                self.wf[word].append([index, fre])
            self.doc_freqs.append(frequencies)
        self.avgdl = float(num_doc) / self.corpus_size
        # collect idf sum to calculate an average idf for epsilon value
        idf_sum = 0
        # collect words with negative idf to set them a special epsilon value.
        # idf can be negative if word is contained in more than half of documents
        negative_idfs = []
        for word, freq in iteritems(nd):
            idf = math.log(self.corpus_size - freq + 0.5) - math.log(freq + 0.5)
            self.idf[word] = idf
            idf_sum += idf
            if idf < 0:
                negative_idfs.append(word)
        self.average_idf = float(idf_sum) / len(self.idf)
        eps = EPSILON * self.average_idf
        for word in negative_idfs:
            self.idf[word] = eps

    def get_score_for_one_word(self, word_one):
        """Calculates score for one word."""
        if word_one not in self.idf.keys():
            return np.zeros(self.corpus_size)
        idf = self.idf[word_one]  # 1*1
        fre_array = np.zeros(self.corpus_size)
        fre_array[np.array(self.wf[word_one])[:, 0]] = np.array(self.wf[word_one])[:, 1]
        corups_arry = fre_array * self.k1 / (fre_array + np.array(self.K)) * idf
        return corups_arry

    def calculate_scores(self, document):
        """Calculates scores for the whole query."""
        # idf_array = np.empty(len(document))  # query_word * 1
        # fre_arrry = np.empty((len(document), self.corpus_size))  # query_word * corpus_size
        # for index, word in enumerate(document):
        #     idf_array[index] = self.idf.get(word, 0)
        #     fre_arrry[index] = self.wf.get(word, [0] * self.corpus_size)
        idf_array = np.array(list(map(lambda x: self.idf.get(x[0], 0), document)))  # query_word * 1
        fre_array = np.zeros((len(document), self.corpus_size))
        for index, item in enumerate(document):
            fre_array[index][np.array(self.wf[item[0]])[:, 0]] = np.array(self.wf[item[0]])[:, 1]
        fre_arrray = fre_array.T  # corpus_size * query_word
        K = np.array([self.K] * len(document)).T
        all_score = np.sum(fre_arrray * self.k1 / (fre_arrray + K) * idf_array, 1)  # corpus_size * 1
        return all_score

    @catch_exception
    def calculate_scores_with_bow(self, document):
        """Calculates scores for the whole query."""
        idf_array = np.array(list(map(lambda x: self.idf.get(x[0], 0), document)))  # query_word * 1
        word_fre_array = np.array(list(map(lambda x: x[1], document)))
        fre_array = np.zeros((len(document), self.corpus_size))
        for index, item in enumerate(document):
            fre_array[index][np.array(self.wf[item[0]])[:, 0]] = np.array(self.wf[item[0]])[:, 1]
        fre_array = fre_array.T  # corpus_size * query_word
        K = np.array([self.K] * len(document)).T
        all_score = np.sum(fre_array * self.k1 / (fre_array + K) * idf_array * word_fre_array, 1)  # corpus_size * 1
        return all_score

    def get_scores(self, document):
        """Computes and returns BM25 scores of given `document` in relation to
        every item in corpus.

        Parameters
        ----------
        document : list of str
            Document to be scored.

        Returns
        -------
        list of float
            BM25 scores.

        """
        # score_array = np.zeros(self.corpus_size)
        # for word in document:
        #     if word not in self.word2score.keys():
        #         # score_word_1 = self.get_scores_ori([word])
        #         score_word = self.get_score_for_one_word(word)
        #         score_array = score_array + score_word
        #         self.word2score[word] = score_word
        #     else:
        #         score_array = score_array + self.word2score[word]
        # return score_array
        return self.calculate_scores_with_bow(document)

    def get_score(self, document, index):
        """Computes BM25 score of given `document` in relation to item of corpus selected by `index`.

        Parameters
        ----------
        document : list of str
            Document to be scored.
        index : int
            Index of document in corpus selected to score with `document`.

        Returns
        -------
        float
            BM25 score.

        """
        score = 0
        doc_freqs = self.doc_freqs[index]
        for word in document:
            if word not in doc_freqs:
                continue
            score += (self.idf[word] * doc_freqs[word] * self.k1
                      / (doc_freqs[word] + self.K[index]))
        return score

    def get_scores_ori(self, document):
        """Computes and returns BM25 scores of given `document` in relation to
        every item in corpus.

        Parameters
        ----------
        document : list of str
            Document to be scored.

        Returns
        -------
        list of float
            BM25 scores.

        """
        scores = np.array([self.get_score(document, index) for index in range(self.corpus_size)])
        return scores

    def get_scores_bow(self, document):
        """Computes and returns BM25 scores of given `document` in relation to
        every item in corpus.

        Parameters
        ----------
        document : list of str
            Document to be scored.

        Returns
        -------
        list of float
            BM25 scores.

        """
        scores = []
        for index in range(self.corpus_size):
            score = self.get_score(document, index)
            if score > 0:
                scores.append((index, score))
        return scores


# def _get_scores_bow(bm25, document):
#     """Helper function for retrieving bm25 scores of given `document` in parallel
#     in relation to every item in corpus.
#
#     Parameters
#     ----------
#     bm25 : BM25 object
#         BM25 object fitted on the corpus where documents are retrieved.
#     document : list of str
#         Document to be scored.
#
#     Returns
#     -------
#     list of (index, float)
#         BM25 scores in a bag of weights format.
#
#     """
#     return bm25.get_scores_bow(document)
#
#
# def _get_scores(bm25, document):
#     """Helper function for retrieving bm25 scores of given `document` in parallel
#     in relation to every item in corpus.
#
#     Parameters
#     ----------
#     bm25 : BM25 object
#         BM25 object fitted on the corpus where documents are retrieved.
#     document : list of str
#         Document to be scored.
#
#     Returns
#     -------
#     list of float
#         BM25 scores.
#
#     """
#     return bm25.get_scores(document)
#
#
# def iter_bm25_bow(corpus, n_jobs=1):
#     """Yield BM25 scores (weights) of documents in corpus.
#     Each document has to be weighted with every document in given corpus.
#
#     Parameters
#     ----------
#     corpus : list of list of str
#         Corpus of documents.
#     n_jobs : int
#         The number of processes to use for computing bm25.
#
#     Yields
#     -------
#     list of (index, float)
#         BM25 scores in bag of weights format.
#
#     Examples
#     --------
#     .. sourcecode:: pycon
#
#         # >>> from gensim.summarization.bm25 import iter_bm25_weights
#         # >>> corpus = [
#         # ...     ["black", "cat", "white", "cat"],
#         # ...     ["cat", "outer", "space"],
#         # ...     ["wag", "dog"]
#         # ... ]
#         # >>> result = iter_bm25_weights(corpus, n_jobs=-1)
#
#     """
#     bm25 = BM25(corpus)
#     import multiprocessing
#     n_processes = multiprocessing.cpu_count() - 1 if multiprocessing.cpu_count() - 1 > n_jobs else n_jobs
#     if n_processes == 1:
#         for doc in corpus:
#             yield bm25.get_scores_bow(doc)
#         return
#
#     get_score = partial(_get_scores_bow, bm25)
#     pool = Pool(n_processes)
#
#     for bow in pool.imap(get_score, corpus):
#         yield bow
#     pool.close()
#     pool.join()
#
#
# def get_bm25_weights(corpus, n_jobs=1):
#     """Returns BM25 scores (weights) of documents in corpus.
#     Each document has to be weighted with every document in given corpus.
#
#     Parameters
#     ----------
#     corpus : list of list of str
#         Corpus of documents.
#     n_jobs : int
#         The number of processes to use for computing bm25.
#
#     Returns
#     -------
#     list of list of float
#         BM25 scores.
#
#     Examples
#     --------
#     .. sourcecode:: pycon
#
#         # >>> from gensim.summarization.bm25 import get_bm25_weights
#         # >>> corpus = [
#         # ...     ["black", "cat", "white", "cat"],
#         # ...     ["cat", "outer", "space"],
#         # ...     ["wag", "dog"]
#         # ... ]
#         # >>> result = get_bm25_weights(corpus, n_jobs=-1)
#
#     """
#     bm25 = BM25(corpus)
#
#     import multiprocessing
#     n_processes = multiprocessing.cpu_count() - 1 if multiprocessing.cpu_count() - 1 > n_jobs else n_jobs
#     if n_processes == 1:
#         weights = [bm25.get_scores(doc) for doc in corpus]
#         return weights
#
#     get_score = partial(_get_scores, bm25)
#     pool = Pool(n_processes)
#     weights = pool.map(get_score, corpus)
#     pool.close()
#     pool.join()
#     return weights
